/* TreeOut.cpp */
/*
()
() **************************************************************************
() *     Tree  TreeInjector             *
() **************************************************************************
()
*/

#include "TreeOut.h"

#include <math.h>

/*
()     : psGraphPrint(...)  (  );
()    :       ...
()                    ;
()   :    ;
()  : ---.
()
*/
void TreePrinter::psGraphPrint( RCTree tree, ostream& os ) const
  {
    String prefix="";
    TreeManipulator treeMan((Tree&)tree);
    while ( treeMan->level() > 1 )
      {
        treeMan.goToParent();
        String oldPrefix=prefix;
        if (  treeMan->parent().childNumber(treeMan.host()) == 0 )
          prefix = " ";
        else
          prefix = "|";
        prefix += oldPrefix;
      }
    os << prefix;

    if ( tree.isRoot() )
      os << "# ";
    else
      {
        os << ( tree.parent().childNumber(tree) ? '|' : '+' );

        if ( tree.hasChild() )
          os << "# ";
        else
          os << String('-') * (tree.root().height()-prefix.length()+2) << "* ";
      }

    os << tree.name();

    if ( tree.isLeaf() && tree.isA() == geneTreeClass )
      os << '[' << RCGeneTree(tree).species() << ']';

    os << endl;

    for ( int i=tree.numberOfChildren()-1; i>=0; i-- )
      psGraphPrint( tree.child(i), os );
  }

/*
()     : phylipPrint(...)  (PHYLIP- ( );
()    :       ...
()                 "-"  ;
()   :    ;
()  : ---.
()
*/
void TreePrinter::phylipPrint( RCBioTree bioTree, ostream& os ) const
  {
    String leftMergin=String("  ")*bioTree.level();

    if ( bioTree.isNotRoot() )
      os << endl << leftMergin;

    if ( bioTree.hasChild() )
      {
        os << '(';
        for ( uint n=0; n<bioTree.numberOfChildren(); n++ )
          {
            phylipPrint( bioTree.child(n), os );
            os << ( (n<bioTree.numberOfChildren()-1) ?  ',' : '\n' );
          }
        os << leftMergin << ')';
      }

    os << bioTree.name();

    if ( bioTree.isNotRoot() )
      os << ':' << bioTree.distanceFromParent();
    else
      os << ';' << endl;
  }

/*
()     : phylipPrint(...)  (PHYLIP- ( );
()    :       ...
()                 "-"  ;
()   :    ;
()  : ---.
()
*/
void TreePrinter::phylipPrint( RCTreeInjection tInject, ostream& os ) const
  {
    String leftMergin=String("  ")*tInject.level();

    if ( tInject.isNotRoot() )
      os << endl << leftMergin;

    if ( tInject.hasChild() )
      {
        os << '(';
        for ( uint n=0; n<tInject.numberOfChildren(); n++ )
          {
            phylipPrint( tInject.child(n), os );
            os << ( (n<tInject.numberOfChildren()-1) ?  ',' : '\n' );
          }
        os << leftMergin << ')';
      }

    os << tInject.name();

    if ( tInject.image().name() != "" )
      os << "--" << tInject.image().name();

    if ( tInject.isRoot() )
      os << ';' << endl;
  }

/*
()     : columnPrint(...)  (  );
()    :       ;
()   : TreeInjector   ;
()  : ---.
()
*/
void TreePrinter::columnPrint( RCTreeInjection injection, ostream& os ) const
  {
//    if ( injection.isLeaf() ) return; //    

    if ( injection.name() != "" )
      os << injection.name();
    else
      os << "";
    if ( injection.hasImage() )
      {
        String nOfPassed(injection.numberOfPassed(),8);
        String nOfCnvrgd(injection.numberOfConverged(),8);
        String tab0(String(' ') * (20-injection.name().length()));
        String tab1(String(' ') * (12-nOfPassed.length()));
        String tab2(String(' ') * (12-nOfCnvrgd.length()));
        os << tab0 << " -( " << nOfPassed << ","
           << tab1 << nOfCnvrgd << tab2 << ")-> ";
        if ( injection.image().name() != "" )
          os << injection.image().name();
        else
          os << "";
      }
    os << endl;

    for ( uint n=0; n<injection.numberOfChildren(); n++ )
      columnPrint( injection.child(n), os );
  }

/*    */
void TreePrinter::printGeneTrees( RCClArray gTrees, ostream& os ) const
  {
    for ( int i=0; i<gTrees.size(); i++ )
//      {
        os << (i+1) << "-:\t" << gTrees[i] << ';' << endl;
//        printGeneTree( RGeneTree(gTrees[i]), os );
//      }
  }

/*    */
void TreePrinter::printTree_Exp( RBioTree bioTree, ostream& os ) const
  {
    if ( GLOBAL_BOOLEAN("ShowPhylipTrees") )
      {
        os << "(*)   PHYLIP-:" << endl;
        phylipPrint( bioTree, os );
      }
    if ( GLOBAL_BOOLEAN("ShowPseudographicsTrees") )
      {
        os << "(*)    :" << endl;
        if ( bioTree.name() == "" )  //     
          bioTree.name("Root");      //    "Root"
        TreeManipulator(bioTree).nameUntitledSubtrees();//  L/R-
        psGraphPrint( bioTree, os );
      }
    if ( GLOBAL_BOOLEAN("ShowTreeInfo") == true )
      os << "(*) :" << endl
         << "       :  " << bioTree.numberOfSubtrees() << ",\n"
         << "      :  " << bioTree.numberOfLeaves() << ",\n"
         << "  () :  " << bioTree.height() << ",\n"
         << " .     :  " << bioTree.maximumPower() << ",\n"
         << "     : ~" << bioTree.sizeInMemory() << " .\n\n";
  }

/*        */
void TreePrinter::printInjections( ClArray gTrees, DuplicationsTree dTree,
                                   ostream& os ) const
  {

    /*      */
    IMPORT_BOOLEAN( showInjections,      "ShowInjections"          );
    IMPORT_BOOLEAN( showInjectionTables, "ShowInjectionTables"     );
    IMPORT_BOOLEAN( showInjectionTrees,  "ShowInjectionTrees"      );
    IMPORT_BOOLEAN( showInjectInfo,      "ShowInjectionsInfo"      );
    IMPORT_BOOLEAN( showInjectStat,      "ShowInjectionStatistics" );

    if ( ! showInjections && ! showInjectInfo && ! showInjectStat )
      return;  // Nothing to do ...

    /*      () */
    for ( int i=0; i<gTrees.size(); i++ )
      {
        RGeneTree gTree=RGeneTree(gTrees[i]);
        if ( gTree.name() == "" )  //     
          gTree.name(String("GeneRoot_")+(i+1));  //   
        TreeManipulator(gTree).nameUntitledSubtrees(); //  L/R-
        if ( GLOBAL_BOOLEAN("NormalizeGeneTrees") )
          gTree.scaleLengths( 1./gTree.numberOfLeaves() );
      }

    /*       () */
    if ( dTree.name() == "" )  //     
      dTree.name("SpecRoot");  //    "SpecRoot"
    TreeManipulator( dTree ).nameUntitledSubtrees(); //  L/R-

    /*   */
    TreeInjector injector(gTrees,dTree);
    const double quality=injector.inject();
    SrtArray injections=injector.injections();

    /*     */
    double meanQuality=0;    //  
    double meanPassed=0;     //   
    double meanConverged=0;  //   

    /*  ,       */
    IMPORT_REAL( costRatio, "CostRatio" );
    IMPORT_BOOLEAN( sortInjections, "SortInjections" );
    for ( int i=0; i<gTrees.size(); i++ )  //    
      {
        RTreeInjection injection =
          ( sortInjections ? RTreeInjection(injections[i]) :
                             RTreeInjection(injector.injections()[i]) );

        if ( sortInjections )
          os << (i+1) << "- ("
             << (injector.injections().firstIndexOf(injections[i])+1)
             << "-) ." << endl;
        else
          os << (i+1) << "- ." << endl;

        if ( showInjections )
          {
            if ( showInjectionTables )
              {
                os << " :" << endl;
                columnPrint( injection, os );
              }
            if ( showInjectionTrees )
              {
                os << "  PHYLIP-:" << endl;
                phylipPrint( injection, os );
              }
          }
        if ( showInjectInfo )
          {
            os << "       : " << injection.quality() << endl;
            os << " ޣ   : " << injection.sumOfPassed()
                 <<  " * " << costRatio << endl;
            os << " ޣ  : " << injection.sumOfConverged() <<endl;
          }
        if ( showInjections || showInjectInfo )
          os << endl;

        /*       */
        meanQuality   += injection.quality();
        meanPassed    += injection.sumOfPassed();
        meanConverged += injection.sumOfConverged();
      }

    if ( !showInjections && !showInjectInfo )
      os << endl;

    os << "   : " << quality << endl;

    if ( !showInjectStat )
      return;  //  ģ 

    /*       */
    dTree.defineNumberOfDuplications( injector );  //  
    os << endl << "  :" << endl << dTree << ';' << endl << endl;

    /*  ޣ   */
    meanQuality   /= gTrees.size();  //
    meanPassed    /= gTrees.size();  //    
    meanConverged /= gTrees.size();  //

    /*     */
    double msdQuality=0;    //   
    double msdPassed=0;     //    
    double msdConverged=0;  //    
    for ( int i=0; i<gTrees.size(); i++ )  //     
      {
        RTreeInjection injection=RTreeInjection(injector.injections()[i]);
        msdQuality   += (injection.quality()        - meanQuality) *
                        (injection.quality()        - meanQuality);
        msdPassed    += (injection.sumOfPassed()    - meanPassed) *
                        (injection.sumOfPassed()    - meanPassed);
        msdConverged += (injection.sumOfConverged() - meanConverged) *
                        (injection.sumOfConverged() - meanConverged);
      }
    /*  ޣ   */
    msdQuality   /= gTrees.size()-1;  //
    msdPassed    /= gTrees.size()-1;  // , ,    (n-1)
    msdConverged /= gTrees.size()-1;  //
    msdQuality   = sqrt(msdQuality);    //
    msdPassed    = sqrt(msdPassed);     //   
    msdConverged = sqrt(msdConverged);  //
    /* ޣ   */

    /*   */
    os << "*****   *****" << endl;
    if ( GLOBAL_BOOLEAN("NormalizeGeneTrees") )
      os << "(  \"\"  .)" << endl;
    os << "   : " << meanQuality   << endl;
    os << "      : " << meanPassed
       <<  " * " << costRatio << endl;
    os << "     : " << meanConverged << endl;
    os << "    : " << msdQuality   << endl;
    os << "       : " << msdPassed
       <<  " * " << costRatio << endl;
    os << "      : " << msdConverged << endl;
    os << endl;

    return;

  }

